/* -*- mode: c++; indent-tabs-mode:nil -*- */
// This file does integrand sampling for Cuba by calling a J verb.
//
// Even though some copying is present the integration is quite efficient
// since many integrand sampling points are calculated by J at once.

#include "jengine.h"
#include <iostream>

// below we mess with Cuba internals to define a new
// sampling routine
typedef struct {
  double *limits;
  jengine *je;
} BackPointers;

typedef /*unsigned*/ int count;
typedef int number;
typedef /*long*/ double Real;
typedef const Real cReal;
typedef struct {
  count ndim, ncomp;
  void* integrand;
  BackPointers* userdata;
  // there is more in Cuba, but here is the necessary minimum
} ShortCubaThis;

const bool TRACE=false;

extern "C" {
  number SampleRaw(ShortCubaThis *t, number m, cReal *x, Real *f) {
    using namespace std;
    if (TRACE) cerr << "sampling "<< m << " points." << endl;
    jengine::monad um=(jengine::monad) t->integrand;
    double* limits=t->userdata->limits;
    jengine* je=t->userdata->je;
    const int n=t->ndim;
    const int k=t->ncomp;
    jarray::I _ttop=je->PROLOG();
    { 
      // now we need to evaluate integrand on m n-dimensional points
      // and obtain m k-dimensional results
      jarray_of_type<double> in(je,n,m);
      // transform coordinates, compute Jacobian and transpose
      jarray_of_type<double> jac(je,m);
      for (int i=0; i<m; i++) {
        jac[i]=1.0;
        for (int j=0; j<n; j++) {
          const double &a=limits[2*j];
          const double &b=limits[2*j+1];
          const double &u=x[i*n+j];
          in(j,i) = a + (b - a)*u;
          jac[i] *= (b - a);
          // The above two lines implement a simple direct mapping
          // between unit hypercube and the volume of integration.
          // The mapping below is more sophisticated, but has a useful
          // feature of weakening potential singularities at the boundary
          // of integration volume. It makes artificial extrema though,
          // which slows down the integration. That's why it is not
          // used.
          // in(j,i) = b*(3-2*u)*u*u + a*(1+2*u)*(u-1)*(u-1);
          // jac[i]*=6*(b-a)*(1-u)*u;
        };
      };
      if (TRACE) cerr << "evaluating integrand at " << in << endl;
      jarray res=um(in);
      if (TRACE) {
        cerr << "integrand returned "<<res.rank()<<"[";
        for(int i=0;i<res.rank();i++) {
          cerr << res.shape()[i];
          if (i<res.rank()-1) cerr << ',';
        };
        cerr<<"] : " << res << endl;
      };
      assert(((res.rank()==2)&&(res.extent(0)==k)&&(res.extent(1)==m))
             || ((k==1)&&(res.rank()==1)&&(res.extent(0)==m)));
      jarray_of_type<double> dres(res);
      if (dres.rank()==2) { // transpose the result
        for (int i=0; i<m; i++)
          for (int j=0; j<k; j++)
            f[i*k+j]=dres(j,i)*jac[i];
      } else {              // just copy vector
        assert(k==1);
        for (int i=0; i<m; i++) f[i]=dres[i]*jac[i];     
      }
    }
    je->EPILOG(_ttop);
    return 0;
  };
}
