/* -*- mode: c++; indent-tabs-mode:nil -*- */
#ifndef JARRAY_H
#define JARRAY_H

#include <vector>
#include <string>
#include <complex>
#include <istream>
#include <ostream>
#include <cassert>
#include "sha1.h"

class jengine;

/**
 * C++ representation of J array. 
 * This is the direct mapping of J array into C++ domain, for support of typed
 * arrays (with automatic type conversion and convenient element indexing) see
 * jarray_of_type template class.
 * Only non-sparse array types are currently supported.
 * This class may operate in standalone mode (managing its memory via malloc), or, if
 * J engine is initalized, it will cooperate with J on memory allocation and make
 * use of J garbage collection (AKA tempstack).
 */
class jarray {
public:
  // C types, corresponding to J types
  // TODO: these are only for 32 bit linux, need ifdefs to cover 
  //       additional platforms
  typedef char B; /**< byte type, equivalent to J */
  typedef char C; /**< literal type, equivalent to J */
  typedef short S; /**< short int type, equivalent to J */
#if _UNIX64
  typedef long long          I;  /**< integer type, equivalent to J */
#else
  typedef long               I;  /**< integer type, equivalent to J */
#endif
  typedef double D; /**< floating point type, equivalent to J */
  /** complex type, equivalent to J */
  typedef struct {
    D re,      /**< real part */
      im;      /**< imaginary part */
  }  Z;

  friend class jengine;

public:
  /** error codes */
  enum errorType {
    ERR_CONV=10,   /**< data type conversion error */
    ERR_SHAPE=11  /**< shape mismatch */
  };

  /** Constants for possible array element types. */
  enum elementType {
    T_B01  =1,           /**< B  boolean                      */
    T_LIT  =2,           /**< C  literal (character)          */
    T_INT  =4,           /**< I  integer                      */
    T_FL   =8,           /**< D  double (IEEE floating point) */
    T_CMPX =16           /**< Z  complex                      */
  };
  // numeric constants, corresponding to J 7.01 types.
  // Not all of J types are supported, but the supported ones have
  // the full set of automatic widening conversions.
  //const static I T_B01  =1;           /**< B  boolean                      */
  //const static I T_LIT  =2;           /**< C  literal (character)          */
  //const static I T_INT  =4;           /**< I  integer                      */
  //const static I T_FL   =8;           /**< D  double (IEEE floating point) */
  //const static I T_CMPX =16;          /**< Z  complex                      */
  //const static I T_BOX  =32;          /**< A  boxed                        */
  //const static I T_XNUM =64;          /**< X  extended precision integer   */
  //const static I T_RAT  =128;         /**< Q  rational number              */
  //const static I T_SB01 =1024;        /**< P  sparse boolean               */
  //const static I T_SLIT =2048;        /**< P  sparse literal (character)   */
  //const static I T_SINT =4096;        /**< P  sparse integer               */
  //const static I T_SFL  =8192;        /**< P  sparse floating point        */
  //const static I T_SCMPX=16384;       /**< P  sparse complex               */
  //const static I T_SBOX =32768;       /**< P  sparse boxed                 */
  //const static I T_SBT  =65536;       /**< SB symbol                       */
  //const static I T_C2T  =131072;      /**< C2 unicode (2-byte characters)  */

protected:
  /** J array header */
  typedef struct {
    I offset,   /**< offset of data w.r.t header. */
      flag,     /**< flags, jarray sets couple of its own. */
      maxbytes, /**< bytes allocated. */
      type,     /**< type of array elements (one of T_XXX) . */
      refcnt,   /**< reference count */
      n,        /**< number of elements in ravel */
      rank,     /**< array rank */
      shape[1]; /**< elements of shape */
  } header;
  header *hdr; /**< Pointer to the array header, NULL for invalid array. */
  jengine *je; /**< Pointer to jengine, managing the memory of this array */

  /** Layout of two words before every array, responsible for J memory
   * management */
  typedef struct {I*a;S j;C mflag,unused;} MS;

  /** Allocates new array in memory.
   * The memory if either malloc-ed (if the first argument is null),
   * or, if passed non-null je pointer, allocated via jtga function 
   * of J engine.
   * @param je_ jengine to manage the memory of this array.
   * @param type type of the new array (one of T_XXXX). 
   * @param rank rank of the new array.
   * @param shape shape of the new array.
   * @return true if success. */
  bool allocate(jengine *je_, elementType type, const I rank, const I* shape);
  
  /*! \cond PRIVATE */
  // conversions: T=bool
  inline static int fromB01(bool& v, const B d) {v=(d!=0); return 0;};
  inline static int fromLIT(bool& v, const C d) {return ERR_CONV;};
  inline static int fromINT(bool& v, const I d) {return ERR_CONV;};
  inline static int fromFL(bool& v, const D d) {return ERR_CONV;};
  inline static int fromCMPX(bool& v, const Z d) {return ERR_CONV;};
  inline static int toB01(const bool v, B& d) {d=v; return 0;};
  inline static int toLIT(const bool v, C& d) {return ERR_CONV;};
  inline static int toINT(const bool v, I& d) {d=v; return 0;};
  inline static int toFL(const bool v,  D& d) {d=v; return 0;};
  inline static int toCMPX(const bool v, Z& d) {d.re=v; d.im=0; return 0;};
  
  // conversions: T=char
  // it may be more convenient to obtain literal array as a string
  // via get(string&) method.
  inline static int fromB01(char& v, const B d) {return ERR_CONV;};
  inline static int fromLIT(char& v, const C d) {v=d; return 0;};
  inline static int fromINT(char& v, const I d) {return ERR_CONV;};
  inline static int fromFL(char& v, const D d) {return ERR_CONV;};
  inline static int fromCMPX(char& v, const Z d) {return ERR_CONV;};
  inline static int toB01(const char v, B& d) {return ERR_CONV;};
  inline static int toLIT(const char v, C& d) {d=v; return 0;};
  inline static int toINT(const char v, I& d) {d=v; return 0;};
  inline static int toFL(const char v,  D& d) {d=v; return 0;};
  inline static int toCMPX(const char v, Z& d) {d.re=v; d.im=0; return 0;};

  // conversions: T=int
  inline static int fromB01(int& v, const B d) {v=d; return 0;};
  inline static int fromLIT(int& v, const C d) {v=d; return 0;};
  inline static int fromINT(int& v, const I d) {v=d; return 0;};
  inline static int fromFL(int& v, const D d) {return ERR_CONV;};
  inline static int fromCMPX(int& v, const Z d) {return ERR_CONV;};
  inline static int toB01(const int v, B& d) {return ERR_CONV;};
  inline static int toLIT(const int v, C& d) {return ERR_CONV;};
  inline static int toINT(const int v, I& d) {d=v; return 0;};
  inline static int toFL(const int v,  D& d) {d=v; return 0;};
  inline static int toCMPX(const int v, Z& d) {d.re=v; d.im=0; return 0;};

  // conversions: T=long
  inline static int fromB01(long& v, const B d) {v=d; return 0;};
  inline static int fromLIT(long& v, const C d) {v=d; return 0;};
  inline static int fromINT(long& v, const I d) {v=d; return 0;};
  inline static int fromFL(long& v, const D d) {return ERR_CONV;};
  inline static int fromCMPX(long& v, const Z d) {return ERR_CONV;};
  inline static int toB01(const long v, B& d) {return ERR_CONV;};
  inline static int toLIT(const long v, C& d) {return ERR_CONV;};
  inline static int toINT(const long v, I& d) {d=v; return 0;};
  inline static int toFL(const long v,  D& d) {d=v; return 0;};
  inline static int toCMPX(const long v, Z& d) {d.re=v; d.im=0; return 0;};

  // conversions: T=long long
  inline static int fromB01(long long& v, const B d) {v=d; return 0;};
  inline static int fromLIT(long long& v, const C d) {v=d; return 0;};
  inline static int fromINT(long long& v, const I d) {v=d; return 0;};
  inline static int fromFL(long long& v, const D d) {return ERR_CONV;};
  inline static int fromCMPX(long long& v, const Z d) {return ERR_CONV;};
  inline static int toB01(const long long v, B& d) {return ERR_CONV;};
  inline static int toLIT(const long long v, C& d) {return ERR_CONV;};
  inline static int toINT(const long long v, I& d) {d=v; return 0;};
  inline static int toFL(const long long v,  D& d) {d=v; return 0;};
  inline static int toCMPX(const long long v, Z& d) {d.re=v; d.im=0; return 0;};

  // conversions: T=double
  inline static int fromB01(double& v, const B d) {v=d; return 0;};
  inline static int fromLIT(double& v, const C d) {v=d; return 0;};
  inline static int fromINT(double& v, const I d) {v=d; return 0;};
  inline static int fromFL(double& v, const D d)  {v=d; return 0;};
  inline static int fromCMPX(double& v, const Z d) {return ERR_CONV;};
  inline static int toB01(const double v, B& d) {return ERR_CONV;};
  inline static int toLIT(const double v, C& d) {return ERR_CONV;};
  inline static int toINT(const double v, I& d) {d=v; return 0;};
  inline static int toFL(const double v,  D& d) {d=v; return 0;};
  inline static int toCMPX(const double v, Z& d) {d.re=v; d.im=0; return 0;};

  // conversions: T=complex<double>
  inline static int fromB01(std::complex<double>& v, const B d) {
    return ERR_CONV;};
  inline static int fromLIT(std::complex<double>& v, const C d) {
    v=d; return 0;};
  inline static int fromINT(std::complex<double>& v, const I d) {
    v=d; return 0;};
  inline static int fromFL(std::complex<double>& v, const D d)  {
    v=d; return 0;};
  inline static int fromCMPX(std::complex<double>& v, const Z d) {
    std::complex<double> c(d.re, d.im); v=c; return 0;};
  inline static int toB01(const std::complex<double> v, B& d) {
    return ERR_CONV;};
  inline static int toLIT(const std::complex<double> v, C& d) {
    return ERR_CONV;};
  inline static int toINT(const std::complex<double> v, I& d) {
    return ERR_CONV;};
  inline static int toFL(const std::complex<double> v,  D& d) {
    return ERR_CONV;};
  inline static int toCMPX(const std::complex<double> v, Z& d) {
    d.re=v.real(); d.im=v.imag(); return 0;};
  /*! \endcond */

  /** Increments refcount */
  void grab() const;
  
  /** Decrements refcount and frees memory, if necessary */
  void release();

  // returns pointer to the header and (if parameter is "true", which is
  // default) marks the array as owned by J preventing its memory from 
  // being freed as it goes out of scope.
  inline I getHeader(bool give_up_ownership=true) const {
    if (give_up_ownership) hdr->flag|=OWNED_BY_J;
    return (I)hdr;
  };

private:
  // flag, marking the array as J owned, preventing its memory from
  // being freed when the array goes out of scope.
  const static int OWNED_BY_J=131072;
  // marks the array as being allocated on heap without J assistance
  const static int OWNED_BY_JPLUS=262144;

public:
  /** returns the array refcount 
   * @return the array refcount. */
  I getRefcount() const {return hdr->refcnt;};

  /** Obtains the value of a specified element of ravel.
   * The type must be convertible. 
   * @param v the place to store the value.
   * @param i zero-based index of the element in ravel.
   * @return 0 on success.
   * */
  template<class T> inline int get(T& v, const int i) const {
    switch(hdr->type) {
    case T_B01: return fromB01(v,((B*)(((B*)hdr)+hdr->offset))[i]);
    case T_LIT: return fromLIT(v,((C*)(((B*)hdr)+hdr->offset))[i]);
    case T_INT: return fromINT(v,((I*)(((B*)hdr)+hdr->offset))[i]);
    case T_FL: return fromFL(v,((D*)(((B*)hdr)+hdr->offset))[i]);
    case T_CMPX: return fromCMPX(v,((Z*)(((B*)hdr)+hdr->offset))[i]);
    default: return ERR_CONV;
    };
  };

  /** Assigns the value of the specified element of ravel.
   * The type must be convertible.
   * @param i zero-based index of the element to assign.
   * @param v the value to put into the array.
   * @return 0 on success. */
  template<class T> inline int set(const int i, const T v) {
    switch(hdr->type) {
    case T_B01: return toB01(v,((B*)(((B*)hdr)+hdr->offset))[i]);
    case T_LIT: return toLIT(v,((C*)(((B*)hdr)+hdr->offset))[i]);
    case T_INT: return toINT(v,((I*)(((B*)hdr)+hdr->offset))[i]);
    case T_FL: return toFL(v,((D*)(((B*)hdr)+hdr->offset))[i]);
    case T_CMPX: return toCMPX(v,((Z*)(((B*)hdr)+hdr->offset))[i]);
    default: return ERR_CONV;
    };
  };

  /** Fits ravel of the array into the specified STL vector. The type
   * must be convertible. Does copy. 
   * @param v the STL vector to hold the copy of the array.
   * @return 0 on success. */
  template<class T> int get(std::vector<T>& v) const {
    std::vector<T> r;
    r.reserve(hdr->n);
    int e;
    switch(hdr->type) {
    case T_B01: for(int i=0;i<hdr->n;i++)
        if (e=fromB01(v[i],((B*)(((B*)hdr)+hdr->offset))[i])) return e;
      break;
    case T_LIT: for(int i=0;i<hdr->n;i++)
        if (e=fromLIT(v[i],((C*)(((B*)hdr)+hdr->offset))[i])) return e;
      break;
    case T_INT: for(int i=0;i<hdr->n;i++)
        if (e=fromINT(v[i],((I*)(((B*)hdr)+hdr->offset))[i])) return e;
      break;
    case T_FL: for(int i=0;i<hdr->n;i++)
        if (e=fromFL(v[i],((D*)(((B*)hdr)+hdr->offset))[i])) return e;
      break;
    case T_CMPX: for(int i=0;i<hdr->n;i++)
        if (e=fromCMPX(v[i],((Z*)(((B*)hdr)+hdr->offset))[i])) return e;
      break;
    default: return ERR_CONV;
    };
    r.swap(v);
    return 0;
  };

  /** Attempts to fit the whole array into the specified type.
   * Currently, this works for literal arrays, which can be 
   * fitted into C++ strings. 
   * @param v place to store the array.
   * @return 0 on success. */
  template<class T> int get(T& v);

  /** Invalid array, a valid array can be assigned to it. */
  jarray();

  /** Loads the array from binary representation in a stream. 
   * The representation can be created by the write method. 
   * @param je_ jengine to own the array's memory.
   * @param in stream to read the array from. */
  jarray(jengine *je_, std::istream& in);

  /** Size (in bytes) of the particular data type.
   * @param type one of T_XXX constants). 
   * @return size of the array element. */
  static jarray::I esize(elementType type);

  /** Size (in bytes) of the single element of this array. */
  jarray::I esize() const;

  /** Add this array (both shape and values) to the hash. 
   * @param sha hash to add the array to. */
  void addhash(SHA1 &sha);

  /** Writes binary representation of this array into 
   * the specified output stream.
   * @param out the stream to write the array to.
   * @return true if successfully written. */
  bool write(std::ostream& out);

  /** Returns true if the array is valid.
   * @return true if the array is valid. */
  inline bool isValid() const {
    return hdr!=NULL;
  };

  /** Returns array type. One of T_XXX constants.
   * @return type of the array elements. */
  inline const I type() const {
    return hdr->type;
  };

  /** Rank (dimensionality) of the array.
   * @return rank of the array. */
  inline const I rank() const {
    return hdr->rank;
  };

  /** Returns pointer to the first element of the shape.
   * @return pointer to the shape. */
  inline I* shape() const {
    return &hdr->shape[0];
  };

  /** Returns the extent of the specified dimension, element of shape.
   * @param dimension the axis, whose size is requested.
   * @return dimension along the specified axis. */
  int extent(int dimension) const {
    assert((dimension>=0)&&(dimension<rank()));
    return shape()[dimension];
  };

  /** Copies array shape into STL vector.
   * @param shape the vector to hold the requested shape.*/
  inline void shape(std::vector<I> &shape) const {
    std::vector<I> r(&hdr->shape[0],&hdr->shape[rank()]);
    r.swap(shape);
  };

  /** Number of elements in ravel. 
   * @return number of elements in ravel. */
  inline const int size() const {
    int n=1;
    for (int i=0;i<rank();i++) n*=hdr->shape[i];
    return n;
  };  

  /** Returns pointer to the beginning of the array data 
   * @return pointer to array data. */
  inline I* data() const {
    return (I*)(((B*)hdr)+hdr->offset);
  };

  /** Returns pointer to jengine, managing memory of this array.
   * @return jengine, menaging the memory of this array, or
   *         NULL if the memory is managed autonomously (via
   *         malloc).
   */
  jengine* getEngine() const {
    return je;
  };

  /** Instantiates on top of an existing J array.
   * @param je_ jengine to own the array memory.
   * @param hdr_ pointer to J array header. */
  jarray(jengine *je_, void *hdr_);

  /** Creates new multidimensional array of the given J type.
   * @param je_ jengine to own the array memory.
   * @param type the type of new array (one of T_XXX).
   * @param rank rank of the new array.
   * @param shape pointer to elements of shape.*/
  jarray(jengine *je_, elementType type, I rank, I* shape);

  /** Creates new multidimensional array of the given J type.
   * @param je_ jengine to own the array memory.
   * @param type the array type (one of T_XXX).
   * @param shape the STL vector, holding elements of shape. */
  jarray(jengine *je_, elementType type, const std::vector<I>& shape);

  /** Creates T_LIT vector from C++ string. 
   * @param je_ jengine to own the array memory.
   * @param str string the new array will contain. */
  jarray(jengine *je_,const std::string &str);

  /** Makes a copy of another array (increments refcount). 
   * @param other the array to copy. */
  jarray(const jarray &other);

  /** Assigns another array to this one (increments refcount and 
   * frees memory, if necessary) 
   * @param other array to assign to this one.
   * @return reference to this array. */
  jarray& assign(const jarray& other);
  
  /** Shortcut to assign 
   * @param other array to assign to this one.
   * @return reference to this array. */
  inline jarray& operator=(const jarray& other) {
    return assign(other);
  };

  /** Performs by-element comparison and return true if two arrays
   *  are the same.
   *  @param rhs array to compare to.
   *  @return true if arrays are the same.
   */
  bool operator== ( const jarray& rhs ) const;

  /** Decrements refcount, frees array memory, if necessary. */
  ~jarray();
};

/** prints ASCII representation of the array
 * @param stream stream to print the array to.
 * @param array the array to print.
 * @return the reference to stream. */
std::ostream& operator<< (std::ostream& stream, const jarray& array);

/** Typed variant of jarray, performs automatic type conversion
 * on instantiation. */
template<class T> class jarray_of_type: public jarray {
private:
  static elementType jtype(I dummy) {return T_INT;};
  static elementType jtype(D dummy) {return T_FL;};
  static elementType jtype(Z dummy) {return T_CMPX;};
  static elementType jtype(std::complex<D> dummy) {return T_CMPX;};
  
public:
  /** Initialize with type conversion.
   * In case no conversion is needed the original array is accessed directly,
   * otherwise a copy is made.
   * To access the array members (subscripting) round backets are used like in
   * Blitz++ and not the square ones, like in Boost. Convenience methods
   * are provided for ranks up to 5, for bigger ranks one has to construct
   * and use for indexing an integer vector. 
   * @param ja jarray, holding the data. */
  jarray_of_type(const jarray &ja):jarray() {
    T dummy;
    elementType type=jtype(dummy);
    if (type!=ja.type()) { // make an independent copy and convert type
      allocate(ja.getEngine(), type, ja.rank(), ja.shape());
      int n=ja.size();
      for (int i=0;i<n;i++) {
        T v;
        ja.get(v,i);
        set(i,v);
      };
    } else assign(ja); // type is the same, direct access
  };

  /** Directly access element of ravel. 
   * @param i index.
   * @return reference to the element. */
  T& operator[](const I i) {
    // assert((i>=0)&&(i<size());
    return ((T*)data())[i];
  };

  /** Direct read-only access to an element of ravel.
   * @param i index.
   * @return constant reference to the element. */
  const T& operator[](const I i) const {
    // assert((i>=0)&&(i<size());
    return ((T*)data())[i];
  };

  /** Conveniently allocate scalar. 
   * @param je_ jengine to manage memory of this array. */
  jarray_of_type(jengine *je_) {
    T dummy; I shape[]={}; allocate(je_, jtype(dummy), 0, shape);
  };

  /** Conveniently allocate 1-D array.
   * @param je_ jengine to manage memory of this array.
   * @param l0 new array dimension */
  jarray_of_type(jengine *je_, const I l0) {
    T dummy; I shape[]={l0}; allocate(je_, jtype(dummy), 1, shape);
  };
  /** conveniently access 1-D array.
   * @param i0 zero-based index of the element to access. 
   * @return reference to the element. */
  T& operator()(const I i0) {
    assert(rank()==1); return ((T*)data())[i0];
  };
  /** Convenient read-only access to 1-D array. 
   * @param i0 zero-based index of the element to access. 
   * @return const reference to the element. */
  const T& operator()(const I i0) const {
    assert(rank()==1); return ((T*)data())[i0];
  };

  /** Conveniently allocate 2-D array.
   * @param je_ jengine to manage memory of this array.
   * @param l0 new array 1-st dimension
   * @param l1 new array 2-nd dimension */
  jarray_of_type(jengine *je_, const I l0, const I l1) {
    T dummy; I shape[]={l0,l1}; allocate(je_,jtype(dummy), 2, shape);
  };

  /** Conveniently access 2-D array.
   * @param i0 zero-based 1-st index of the element to access.
   * @param i1 zero-based 2-nd index of the element to access. 
   * @return reference to the element. */
  T& operator()(const I i0,const I i1) {
    assert(rank()==2); return ((T*)data())[i0*extent(1)+i1];
  };
  /** Convenient read-only access to 2-D array.
   * @param i0 zero-based 1-st index of the element to access.
   * @param i1 zero-based 2-nd index of the element to access. 
   * @return const reference to the element. */
  const T& operator()(const I i0,const I i1) const {
    assert(rank()==2); return ((T*)data())[i0*extent(1)+i1];
  };
  
  /** Conveniently allocate 3-D array.
   * @param je_ jengine to manage memory of this array.
   * @param l0 new array 1-st dimension
   * @param l1 new array 2-nd dimension 
   * @param l2 new array 3-rd dimension */
  jarray_of_type(jengine *je_, const I l0, const I l1, const I l2) {
    T dummy; I shape[]={l0,l1,l2}; allocate(je_, jtype(dummy), 3, shape);
  };

  /** Conveniently access 3-D array.
   * @param i0 zero-based 1-st index of the element to access.
   * @param i1 zero-based 2-nd index of the element to access. 
   * @param i2 zero-based 3-rd index of the element to access. 
   * @return reference to the element. */
  T& operator()(const I i0,const I i1,const I i2) {
    assert(rank()==3); return ((T*)data())[(i0*extent(1)+i1)*extent(2)+i2];
  };

  /** Convenient read-only access to 3-D array.
   * @param i0 zero-based 1-st index of the element to access.
   * @param i1 zero-based 2-nd index of the element to access. 
   * @param i2 zero-based 3-rd index of the element to access. 
   * @return const reference to the element. */
  const T& operator()(const I i0,const I i1,const I i2) const {
    assert(rank()==3); return ((T*)data())[(i0*extent(1)+i1)*extent(2)+i2];
  };

  /** Conveniently allocate 4-D array.
   * @param je_ jengine to manage memory of this array.
   * @param l0 new array 1-st dimension
   * @param l1 new array 2-nd dimension 
   * @param l2 new array 3-rd dimension 
   * @param l3 new array 4-th dimension */
  jarray_of_type(jengine *je_, const I l0, const I l1, const I l2, 
                 const I l3) {
    T dummy; I shape[]={l0,l1,l2,l3}; allocate(je_,jtype(dummy), 4, shape);
  };

  /** Conveniently access 4-D array.
   * @param i0 zero-based 1-st index of the element to access.
   * @param i1 zero-based 2-nd index of the element to access. 
   * @param i2 zero-based 3-rd index of the element to access. 
   * @param i3 zero-based 4-th index of the element to access. 
   * @return reference to the element. */
  T& operator()(const I i0,const I i1,const I i2,const I i3) {
    assert(rank()==4);
    return ((T*)data())[((i0*extent(1)+i1)*extent(2)+i2)*extent(3)+i3];
  };

  /** Convenient read-only access to 4-D array.
   * @param i0 zero-based 1-st index of the element to access.
   * @param i1 zero-based 2-nd index of the element to access. 
   * @param i2 zero-based 3-rd index of the element to access. 
   * @param i3 zero-based 4-th index of the element to access. 
   * @return const reference to the element. */
  const T& operator()(const I i0,const I i1,const I i2,const I i3) const {
    assert(rank()==4);
    return ((T*)data())[((i0*extent(1)+i1)*extent(2)+i2)*extent(3)+i3];
  };

  /** Conveniently allocate 5-D array.
   * @param je_ jengine to manage memory of this array.
   * @param l0 new array 1-st dimension
   * @param l1 new array 2-nd dimension 
   * @param l2 new array 3-rd dimension 
   * @param l3 new array 4-th dimension
   * @param l4 new array 5-th dimension */
  jarray_of_type(jengine *je_, const I l0, const I l1, const I l2, 
                 const I l3, const I l4) {
    T dummy; I shape[]={l0,l1,l2,l3,l4}; allocate(je_,jtype(dummy), 5, shape);
  };

  /** Conveniently access 5-D array.
   * @param i0 zero-based 1-st index of the element to access.
   * @param i1 zero-based 2-nd index of the element to access. 
   * @param i2 zero-based 3-rd index of the element to access. 
   * @param i3 zero-based 4-th index of the element to access.
   * @param i4 zero-based 5-th index of the element to access. 
   * @return reference to the element. */
  T& operator()(const I i0,const I i1,const I i2,const I i3,const I i4) {
    assert(rank()==4);
    return ((T*)data())[(((i0*extent(1)+i1)*extent(2)+i2)*extent(3)+i3)*extent(4)+i4];
  };

  /** Convenient read-only access to 5-D array.
   * @param i0 zero-based 1-st index of the element to access.
   * @param i1 zero-based 2-nd index of the element to access. 
   * @param i2 zero-based 3-rd index of the element to access. 
   * @param i3 zero-based 4-th index of the element to access.
   * @param i4 zero-based 5-th index of the element to access. 
   * @return const reference to the element. */
  const T& operator()(const I i0,const I i1,const I i2,const I i3,const I i4) const {
    assert(rank()==4);
    return ((T*)data())[(((i0*extent(1)+i1)*extent(2)+i2)*extent(3)+i3)*extent(4)+i4];
  };
  

  /** Conveniently allocate n-D array.
   * @param je_ jengine to manage memory of this array.
   * @param rank rank of new array.
   * @param shape shape of new array. */
  jarray_of_type(jengine *je_, const I rank, const I *shape) {
    T dummy; allocate(je_, jtype(dummy), rank, shape);
  };

  /** Conveniently allocate n-D array.
   * @param je_ jengine to manage memory of this array.
   * @param shape shape of new array. */
  jarray_of_type(jengine *je_, const std::vector<I> &shape) {
    T dummy; allocate(je_, jtype(dummy), shape.size(), &shape[0]);
  };

  /** Conveniently access n-D array.
   * @param subscripts subscripts of the element to access.
   * @return reference to the element. */
  T& operator()(const std::vector<I> &subscripts) {
    assert(rank()==subscripts.size());
    I pos=0;
    for(int i=0;i<subscripts.size();i++) pos=pos*extent(i)+subscripts[i];
    return ((T*)data())[pos];
  };
  /** Convenient read-only access to n-D array.
   * @param subscripts subscripts of the element to access.
   * @return const reference to the element. */
  const T& operator()(const std::vector<I> &subscripts) const {
    assert(rank()==subscripts.size());
    I pos=0;
    for(int i=0;i<subscripts.size();i++) pos=pos*extent(i)+subscripts[i];
    return ((T*)data())[pos];
  };
  /** Conveniently access n-D array.
   * Be careful, allocating enough subscripts.
   * @param subscripts subscripts of the element to access.
   * @return reference to the element. */
  T& operator()(const I *subscripts) {
    I pos=0;
    for(int i=0;i<rank();i++) pos=pos*extent(i)+subscripts[i];
    return ((T*)data())[pos];
  };
  /** Convenient read-only access to n-D array.
   * Be careful, allocating enough subscripts.
   * @param subscripts subscripts of the element to access.
   * @return const reference to the element. */
  const T& operator()(const I *subscripts) const {
    I pos=0;
    for(int i=0;i<rank();i++) pos=pos*extent(i)+subscripts[i];
    return ((T*)data())[pos];
  };
  
};
#endif
