/* -*- mode: c++; indent-tabs-mode:nil -*- */
// verbs for performing various Fourier transforms using fftw library

#include <jarray.h>
#include <fftw3.h>
#include <complex>
#include <vector>
#include <iostream>

using namespace std;

// complex to complex fft
// left argument : scalar sign (boolean: 0 -- forward fft, 1 -- back),
// right argument: data vector (complex)
// produces      : fft complex vector
jarray fft(jarray x, jarray y) {
  assert(x.getEngine()==y.getEngine());
  jengine *je=y.getEngine();
  // TODO: take care of the array alignment for SIMD
  bool sign;
  x.get(sign,0); // left argument must be a boolean scalar
  jarray_of_type< complex<double> > data(y);
  const int r=y.rank();
  vector<jarray::I> shape;
  y.shape(shape);
  jarray_of_type< complex<double> > fft(je, shape);
  fftw_plan p;
  int shapeI[r]; for(int i=0;i<r;i++) shapeI[i]=shape[i]; // convert type
  // assuming FFTW_ESTIMATE does not overwrite arrays
  p=fftw_plan_dft(r, shapeI, (fftw_complex*)&data[0], 
                  (fftw_complex*)&fft[0], 
                  sign?FFTW_BACKWARD:FFTW_FORWARD,
                  FFTW_ESTIMATE);
  fftw_execute(p);
  fftw_destroy_plan(p);
  // TODO: study how significant will be performance improvement from
  //       additional (to what fftw already does) plan caching.
  return fft;
};
jarray fft(jarray y) {
  jengine *je=y.getEngine();
  std::vector<jarray::I> shape;
  jarray mul(je, jarray::T_B01, shape);
  mul.set(0,false);
  return fft(mul,y);
};

// monad: forward real to complex fft
// argument: data vector (real)
// produces: fft complex vector
jarray fftr2c(jarray y) {
  jengine *je=y.getEngine();
  // TODO: take care of the array alignment for SIMD
  jarray_of_type< double > data(y);
  const int r=y.rank();
  vector<jarray::I> shape;
  y.shape(shape);
  shape[r-1]=shape[r-1]/2+1;
  jarray_of_type< complex<double> > fft(je,shape);
  fftw_plan p;
  y.shape(shape);
  int shapeI[r]; for(int i=0;i<r;i++) shapeI[i]=shape[i]; // convert type
  // assuming FFTW_ESTIMATE does not overwrite arrays
  p=fftw_plan_dft_r2c(r, shapeI, &data[0], (fftw_complex*)&fft[0],
                      FFTW_ESTIMATE);
  fftw_execute(p);
  fftw_destroy_plan(p);
  // TODO: study how significant will be performance improvement from
  //       additional (to what fftw already does) plan caching.
  return fft;
};

// dyad: backward complex to real
// left argument : logical size of the array (shape of the original real data)
// right argument: fft complex vector
// produces      : data vector (real)
jarray fftc2r(jarray x, jarray y) {
  // TODO: take care of the array alignment for SIMD
  assert(x.getEngine()==y.getEngine());
  jengine *je=y.getEngine();
  jarray_of_type<jarray::I> shape(x);
  const int r=x.size();
  assert(((x.rank()==1)&&(x.size()==y.rank()))
         || ((y.rank()==1) && (x.rank()==0) && (r==1)));
  for (int i=0;i<r-1;i++) assert(shape[i]==y.shape()[i]);
  assert(y.shape()[r-1]==(shape[r-1]/2+1)); // last dimension is halved
  jarray_of_type< complex<double> > fft(y);
  vector<jarray::I> dshape(r);
  for (int i=0; i<r; i++) dshape[i]=shape[i];
  jarray_of_type< double > data(je, dshape);
  fftw_plan p;
  int shapeI[r]; for(int i=0;i<r;i++) shapeI[i]=dshape[i]; // convert type
  // assuming FFTW_ESTIMATE does not overwrite arrays
  p=fftw_plan_dft_c2r(r, shapeI, (fftw_complex*)&fft[0], &data[0], 
                      FFTW_ESTIMATE);
  fftw_execute(p);
  fftw_destroy_plan(p);
  // TODO: study how significant will be performance improvement from
  //       additional (to what fftw already does) plan caching.
  return data;
};
