/* -*- mode: c++; indent-tabs-mode:nil -*- */
#include <fstream>
#include <string>

/** YACTS trajectory file. Holds the past and current states of integrated system. */
class trjfile {
protected:
  /** Number of equations in the system solved (frame size). */
  int neq;
  /** Opened stream to the underlying file. */
  std::fstream trj;
public:
  
  /** Instantiates unopened trajectory file. */
  trjfile();

  /** Opens/creates trajectory and positions it before the last frame, or right after
      the header if trajectory was empty.
      * @param path filesystem path to the trajectory. 
      * @param create when true an empty trajectory will be created 
      *        when file is missing, otherwise this is failure. 
      * @return true if everything is ok. */
  bool open(std::string path, bool create=true);
  
  /** Closes the trajectory */
  bool close();

  /** Checks that trajectory file is opened and properly positioned.
   * @return true if everything is ok. */
  bool good();

  /** Saves frame to the file at the current pos, appends if necessary. 
   * If this is the first frame saved, the passed neq parameter _defines_
   * the number of frames in this trajectory file, then it becomes an error
   * to save frames with different neq.
   * @param neq number of equations in the system.
   * @param T time of the frame. 
   * @param data frame data. 
   * @return true if success. */
  bool saveFrame(const int neq, const double &T, const double* data);
  
  /** Checks if the file has next frame stored.
   * @return true if the next frame can be read. */
  bool hasNextFrame();

  /** Get number of equations.
   * @return the number of frames or 0 if the file was just created and
   *         no frames were stored yet. */
  int getNEQ();

  /** Load frame.
   * @param T the variable to hold the time of the frame.
   * @param data the array to hold the frame data (should have
   *        enough space for getNEQ() doubles). */
  bool loadFrame(double &T, double* data);

  /** Get number of frames in the file.
   * @return the total number of frames currently in trajectory. */
  int size();

  /** Position the file before the specified frame.
   * @param frame position will be set before this frame, 0 positions
   *        before the first frame, -1 positions before the last frame. */
  bool toFrame(int frame);

protected:
  /** header size on disk */
  std::streampos header_size();
  
  /** frame size on disk */
  std::streampos frame_size();

  /** patch the header with new data 
   * @param id reserved.
   * @param neq number of equations in the system. */
  void setHeader(int id, int neq);

};
